package org.netkernel.layer0.logging;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.net.URL;
import java.util.List;
import java.util.logging.Formatter;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.netkernel.container.ILogger;
import org.netkernel.container.IMessages;
import org.netkernel.layer0.util.FastSchematron;
import org.netkernel.layer0.util.PairList;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLReadable;
import org.netkernel.layer0.util.XMLUtils;
import org.w3c.dom.*;

public class ApplicationLogImpl implements IApplicationLog
{
	private final Logger mLogger;
	private final LogManager mManager;
	private final String mName;
	private final boolean[] mShouldLog;
	private String mStandardLogFile;
	
	private static final PairList VALIDATION = new PairList(4);
	static
	{	VALIDATION.put("/*[name()!='log']", "MSG_LR0_LOG_ROOT");
		VALIDATION.put("/log/*[name()!='id' and name()!='name' and name()!='handlers']", "MSG_LR0_LOG_ROOT_TAGS");
		VALIDATION.put("/log[string-length(id)=0]", "MSG_LR0_LOG_ID");
		VALIDATION.put("/log[string-length(name)=0]", "MSG_LR0_LOG_NAME");
		VALIDATION.put("/log[not(handlers)]", "MSG_LR0_LOG_HANDLERS");
		VALIDATION.put("/log/handlers/*[name()!='handler']", "MSG_LR0_LOG_HANDLER_TAGS");
		VALIDATION.put("/log/handlers[not(handler)]", "MSG_LR0_LOG_HANDLER");
		VALIDATION.put("/log/handlers/handler/*[name()!='instance' and name()!='formatterClass']", "MSG_LR0_LOG_HANDLER_DETAILS");
	}
	
	public ApplicationLogImpl(URL aConfigResource, IMessages aMessages, LogManager aManager) throws Exception
	{	mManager=aManager;
		final int N=8;
		mShouldLog=new boolean[N];
		for (int i=0; i<N; i++) mShouldLog[i]=true;
		XMLReadable config = new XMLReadable(aConfigResource.openStream());
		Element root=((Document)config.getContextNode()).getDocumentElement();
		PairList validation=FastSchematron.validate(root, VALIDATION);
		if (validation.size()>0)
		{	String m = FastSchematron.formatValidationError(validation,aMessages);		
			throw Utils.createFormattedException("EX_LR0_LOG_CONFIG", "MSG_RAW", aMessages, m);
		}		
		
		String id=config.getText("/log/id");
		mLogger=Logger.getLogger(id);
		mLogger.setUseParentHandlers(false);
		mLogger.setLevel(Level.ALL);
		for (Handler h:mLogger.getHandlers())
		{	h.close();
			mLogger.removeHandler(h);
		}
		
		mName=config.getText("/log/name");
		
		List<Node> handlerNodes=config.getNodes("/log/handlers/handler");
		for (Node handlerNode : handlerNodes)
		{
			XMLReadable handlerConfig=new XMLReadable(handlerNode);
			XMLReadable instance=new XMLReadable(handlerConfig.getFirstNode("instance"));
			Handler h=(Handler)constructFromXML(instance);
			h.setLevel(Level.ALL);
			String classname=handlerConfig.getText("formatterClass");
			Formatter f=null;
			if(!classname.equals(""))
			{	Class c=Class.forName(classname);
				f=(Formatter)c.newInstance();
				h.setFormatter(f);
			}
			mLogger.addHandler(h);
			
			//determine if it contains a standard handler that is supported by the logviewer
			if (h instanceof LogFileHandler && f instanceof XMLFormatter)
			{	mStandardLogFile=((LogFileHandler)h).getPrimaryLogFile();
			}
		}
	}
	
	public String getName()
	{	return mName;
	}
	
	public String getId()
	{	return mLogger.getName();
	}
	
	public String getStandardLogFile()
	{	return mStandardLogFile;
	}
	
	private static Object constructFromXML(XMLReadable aConfig) throws ClassNotFoundException, NoSuchMethodException, InstantiationException, IllegalAccessException, InvocationTargetException
	{	String classname=null;
		List nodes=aConfig.getNodes("*");
		Class[] classes=new Class[nodes.size()-1];
		Object[] args=new Object[nodes.size()-1];
		int j=0;
		for (int i=0; i<nodes.size(); i++)
		{	Node n=(Node)nodes.get(i);
			String name=n.getNodeName();
			String value=XMLUtils.getText(n);
			if (name.equals("class"))
			{	classname=value;
			}
			else
			{	if (name.equals("string"))
				{	classes[j]=String.class;
					args[j]=value;
				}
				else if (name.equals("int"))
				{	classes[j]=int.class;
					args[j]=new Integer(value);
				}
				else if (name.equals("long"))
				{	classes[j]=long.class;
					args[j]=new Long(value);
				}
				else if (name.equals("boolean"))
				{	classes[j]=boolean.class;
					args[j]=Boolean.valueOf(value);
				}
				else
				{	throw new IllegalArgumentException("Unknown constructor type "+name);
				}
				j++;
			}
		}
		if (classname==null || j!=nodes.size()-1)
		{	throw new IllegalArgumentException("Class must be specified once");
		}
		Class c = Class.forName(classname);
		Constructor cs=c.getConstructor(classes);
		Object instance=cs.newInstance(args);
		return instance;
	}
	
	public boolean shouldLog(int aLevel, Object aCallingObject)
	{	boolean result=false;
		if (aLevel>=0 && aLevel<mShouldLog.length)
		{	result=mShouldLog[aLevel];
		}
		return result;
	}
	
	public void setShouldLog(int aLevel, Object aCallingObject, boolean aShouldLog)
	{	if (aLevel>=0 && aLevel<mShouldLog.length)
		{	mShouldLog[aLevel]=aShouldLog;
		}
	}
	
	
	public void logRaw(int aLevel, Object aCallingObject, String aMessage)
	{	if (shouldLog(aLevel,aCallingObject))
		{	String src;
			if (aCallingObject instanceof String)
			{	src=(String)aCallingObject;
			}
			else if (aCallingObject instanceof Class)
			{	src=((Class)aCallingObject).getName();
				int i=src.lastIndexOf('.');
				if (i>0) src=src.substring(i+1);
			}
			else
			{	src=aCallingObject.getClass().getName();
				int i=src.lastIndexOf('.');
				if (i>0) src=src.substring(i+1);
			}
			mLogger.logp(getLevel(aLevel),src,null,aMessage);
		}
	}
	
	private Level getLevel(int aLevel)
	{
		Level level;
		switch (aLevel)
		{	case ILogger.SEVERE:
				level = Level.SEVERE; break;
			case ILogger.WARNING: 
				level = Level.WARNING; break;
			case ILogger.INFO:
				level = Level.INFO; break;
			case ILogger.FINE:
				level = Level.FINE; break;
			case ILogger.FINER:
				level = Level.FINER; break;
			case ILogger.DEBUG:
				level = Level.FINEST; break;
			case ILogger.CACHE:
				level = Level.FINEST; break;
			default:
				level = Level.SEVERE; break;
		}
		return level;
	}

	public boolean isExpired()
	{	return false;
	}

}
